from datetime import datetime, timedelta
from zoneinfo import ZoneInfo

import time_machine
from dateutil.relativedelta import relativedelta
from django.conf import settings
from django.test import TestCase, override_settings
from edc_protocol.research_protocol_config import ResearchProtocolConfig
from edc_utils import get_utcnow
from faker import Faker
from model_bakery import baker

from consent_app.models import SubjectConsent
from edc_consent.site_consents import site_consents

from ..consent_test_utils import consent_factory

fake = Faker()


@time_machine.travel(datetime(2019, 4, 1, 8, 00, tzinfo=ZoneInfo("UTC")))
@override_settings(
    EDC_PROTOCOL_STUDY_OPEN_DATETIME=get_utcnow() - relativedelta(years=5),
    EDC_PROTOCOL_STUDY_CLOSE_DATETIME=get_utcnow() + relativedelta(years=1),
    EDC_AUTH_SKIP_SITE_AUTHS=True,
    EDC_AUTH_SKIP_AUTH_UPDATER=False,
)
class TestConsentModel(TestCase):
    def setUp(self):
        self.study_open_datetime = ResearchProtocolConfig().study_open_datetime
        self.study_close_datetime = ResearchProtocolConfig().study_close_datetime
        site_consents.registry = {}
        self.consent_v1 = consent_factory(
            proxy_model="consent_app.subjectconsentv1",
            start=self.study_open_datetime,
            end=self.study_open_datetime + timedelta(days=50),
            version="1.0",
        )
        self.consent_v2 = consent_factory(
            proxy_model="consent_app.subjectconsentv2",
            start=self.study_open_datetime + timedelta(days=51),
            end=self.study_open_datetime + timedelta(days=100),
            version="2.0",
        )

        self.consent_v3_start_date = self.study_open_datetime + timedelta(days=101)
        self.consent_v3 = consent_factory(
            proxy_model="consent_app.subjectconsentv3",
            start=self.study_open_datetime + timedelta(days=101),
            end=self.study_open_datetime + timedelta(days=150),
            version="3.0",
            updates=self.consent_v2,
        )
        site_consents.register(self.consent_v1)
        site_consents.register(self.consent_v2, updated_by=self.consent_v3)
        site_consents.register(self.consent_v3)
        self.dob = self.study_open_datetime - relativedelta(years=25)

        self.subject_identifier = "123456789"
        self.identity = "987654321"

        # travel to consent v1 validity period and consent subject
        traveller = time_machine.travel(self.study_open_datetime)
        traveller.start()
        self.v1_consent_datetime = get_utcnow()
        cdef = site_consents.get_consent_definition(report_datetime=self.v1_consent_datetime)
        baker.make_recipe(
            cdef.model,
            subject_identifier=self.subject_identifier,
            identity=self.identity,
            confirm_identity=self.identity,
            consent_datetime=self.v1_consent_datetime,
            dob=get_utcnow() - relativedelta(years=25),
        )
        traveller.stop()

        # travel to consent v2 validity period and consent subject
        traveller = time_machine.travel(self.study_open_datetime + timedelta(days=51))
        traveller.start()
        self.v2_consent_datetime = get_utcnow()
        cdef = site_consents.get_consent_definition(report_datetime=self.v2_consent_datetime)
        baker.make_recipe(
            cdef.model,
            subject_identifier=self.subject_identifier,
            identity=self.identity,
            confirm_identity=self.identity,
            consent_datetime=self.v2_consent_datetime,
            dob=get_utcnow() - relativedelta(years=25),
        )
        traveller.stop()

        # travel to consent v3 validity period and consent subject
        # not this is 10 days into v3 validity period
        traveller = time_machine.travel(cdef.end + relativedelta(days=10))
        traveller.start()
        self.v3_consent_datetime = get_utcnow()
        cdef = site_consents.get_consent_definition(report_datetime=self.v3_consent_datetime)
        baker.make_recipe(
            cdef.model,
            subject_identifier=self.subject_identifier,
            identity=self.identity,
            confirm_identity=self.identity,
            consent_datetime=self.v3_consent_datetime,
            dob=get_utcnow() - relativedelta(years=25),
        )
        traveller.stop()

        # Note: subject now has three consents in the table.
        self.assertEqual(SubjectConsent.objects.filter(identity=self.identity).count(), 3)

    def test_is_v2_within_v2_consent_period(self):
        consent = site_consents.get_consent_or_raise(
            subject_identifier=self.subject_identifier,
            report_datetime=self.consent_v3_start_date - relativedelta(days=5),
            site_id=settings.SITE_ID,
        )
        self.assertEqual(consent.version, "2.0")

    def test_is_v2_after_v2_end_date_but_before_v3_consent_date(self):
        """Assert returns v2 in the gap between the end of the
        v2 consent period and the subject's v3 consent date.
        """
        cdef = site_consents.get_consent_definition(report_datetime=self.v3_consent_datetime)
        SubjectConsent.objects.filter(consent_datetime__range=[cdef.start, cdef.end])

        consent = site_consents.get_consent_or_raise(
            subject_identifier=self.subject_identifier,
            report_datetime=self.v3_consent_datetime - relativedelta(days=5),
            site_id=settings.SITE_ID,
        )
        self.assertEqual(consent.version, "2.0")

    def test_is_v3_on_v3_consent_date(self):
        consent = site_consents.get_consent_or_raise(
            subject_identifier=self.subject_identifier,
            report_datetime=self.v3_consent_datetime,
            site_id=settings.SITE_ID,
        )
        self.assertEqual(consent.version, "3.0")

    def test_is_v3_on_after_v3_consent_date(self):
        consent = site_consents.get_consent_or_raise(
            subject_identifier=self.subject_identifier,
            report_datetime=self.v3_consent_datetime + relativedelta(days=5),
            site_id=settings.SITE_ID,
        )
        self.assertEqual(consent.version, "3.0")



from __future__ import annotations

from typing import TYPE_CHECKING

from django import forms
from edc_constants.constants import NO, YES
from edc_screening.utils import (
    get_subject_screening_model_cls,
    get_subject_screening_or_raise,
)

if TYPE_CHECKING:
    from edc_screening.model_mixins import ScreeningModelMixin


class ConsentModelFormMixinError(Exception):
    pass


class CleanFieldsModelFormValidationMixin:
    """A model form mixin calling the default `clean_xxxxx` django
    methods.

    Used by ConsentModelFormMixin.

    See also: ConsentModelFormValidationMixin
    """

    @property
    def subject_screening_model_cls(self) -> ScreeningModelMixin:
        return get_subject_screening_model_cls()

    @property
    def subject_screening(self):
        screening_identifier = self.cleaned_data.get(
            "screening_identifier"
        ) or self.initial.get("screening_identifier")
        if not screening_identifier:
            raise forms.ValidationError(
                "Unable to determine the screening identifier. "
                f"This should be part of the initial form data. Got {self.cleaned_data}"
            )
        return get_subject_screening_or_raise(screening_identifier, is_modelform=True)

    def clean_consent_reviewed(self) -> str:
        consent_reviewed = self.cleaned_data.get("consent_reviewed")
        if consent_reviewed != YES:
            raise forms.ValidationError(
                "Complete this part of the informed consent process before continuing.",
                code="invalid",
            )
        return consent_reviewed

    def clean_study_questions(self) -> str:
        study_questions = self.cleaned_data.get("study_questions")
        if study_questions != YES:
            raise forms.ValidationError(
                "Complete this part of the informed consent process before continuing.",
                code="invalid",
            )
        return study_questions

    def clean_assessment_score(self) -> str:
        assessment_score = self.cleaned_data.get("assessment_score")
        if assessment_score != YES:
            raise forms.ValidationError(
                "Complete this part of the informed consent process before continuing.",
                code="invalid",
            )
        return assessment_score

    def clean_consent_copy(self) -> str:
        consent_copy = self.cleaned_data.get("consent_copy")
        if consent_copy == NO:
            raise forms.ValidationError(
                "Complete this part of the informed consent process before continuing.",
                code="invalid",
            )
        return consent_copy

    def clean_consent_signature(self) -> str:
        consent_signature = self.cleaned_data.get("consent_signature")
        if consent_signature != YES:
            raise forms.ValidationError(
                "Complete this part of the informed consent process before continuing.",
                code="invalid",
            )
        return consent_signature

    def clean_initials(self) -> str:
        initials = self.cleaned_data.get("initials")
        if initials and initials != self.subject_screening.initials:
            raise forms.ValidationError(
                "Initials do not match those submitted at screening. "
                f"Expected {self.subject_screening.initials}."
            )
        return initials



from django.utils.translation import gettext as _

REQUIRES_CONSENT_FIELDS = (
    "consent_model",
    "consent_version",
)
requires_consent_fieldset_tuple = (
    _("Consent"),
    {"classes": ("collapse",), "fields": REQUIRES_CONSENT_FIELDS},
)



from __future__ import annotations

from datetime import datetime

from edc_protocol.research_protocol_config import ResearchProtocolConfig

from edc_consent.consent_definition import ConsentDefinition


def consent_definition_factory(
    model: str | None = None,
    start: datetime = None,
    end: datetime = None,
    gender: list[str] | None = None,
    updates: ConsentDefinition = None,
    version: str | None = None,
    age_min: int | None = None,
    age_max: int | None = None,
    age_is_adult: int | None = None,
    **kwargs,
) -> ConsentDefinition:
    options = dict(
        start=start or ResearchProtocolConfig().study_open_datetime,
        end=end or ResearchProtocolConfig().study_close_datetime,
        gender=gender or ["M", "F"],
        updates=updates or None,
        version=version or "1",
        age_min=age_min or 16,
        age_max=age_max or 64,
        age_is_adult=age_is_adult or 18,
    )
    options.update(**kwargs)
    model = model or "consent_app.subjectconsentv1"
    consent_definition = ConsentDefinition(model, **options)
    return consent_definition


def consent_factory(proxy_model=None, **kwargs):
    options = dict(
        start=kwargs.get("start"),
        end=kwargs.get("end"),
        gender=kwargs.get("gender", ["M", "F"]),
        updates=kwargs.get("updates", None),
        version=kwargs.get("version", "1"),
        age_min=kwargs.get("age_min", 16),
        age_max=kwargs.get("age_max", 64),
        age_is_adult=kwargs.get("age_is_adult", 18),
        validate_duration_overlap_by_model=kwargs.get(
            "validate_duration_overlap_by_model", True
        ),
    )
    consent_definition = ConsentDefinition(proxy_model, **options)
    return consent_definition



from __future__ import annotations

import sys
from copy import deepcopy
from datetime import datetime
from typing import TYPE_CHECKING

from django.apps import apps as django_apps
from django.core.management.color import color_style
from django.utils.module_loading import import_module, module_has_submodule
from edc_sites.site import sites as site_sites
from edc_utils import ceil_secs, floor_secs, formatted_date, to_utc

from .exceptions import (
    AlreadyRegistered,
    ConsentDefinitionDoesNotExist,
    ConsentDefinitionError,
    ConsentDefinitionNotConfiguredForUpdate,
    SiteConsentError,
)

if TYPE_CHECKING:
    from edc_sites.single_site import SingleSite

    from .consent_definition import ConsentDefinition
    from .stubs import ConsentLikeModel


__all__ = ["site_consents"]


class SiteConsents:
    def __init__(self):
        self.registry = {}
        self.loaded = False

    def register(
        self, cdef: ConsentDefinition, updated_by: ConsentDefinition | None = None
    ) -> None:
        cdef.updated_by = updated_by
        if cdef.name in self.registry:
            raise AlreadyRegistered(f"Consent definition already registered. Got {cdef.name}.")
        self.validate_period_overlap_or_raise(cdef)
        self.validate_updates_or_raise(cdef)
        self.registry.update({cdef.name: cdef})
        self.loaded = True

    def unregister(self, cdef: ConsentDefinition) -> None:
        self.registry.pop(cdef.name, None)

    def get_registry_display(self):
        cdefs = sorted(list(self.registry.values()), key=lambda x: x.version)
        return "', '".join([cdef.display_name for cdef in cdefs])

    def get(self, name) -> ConsentDefinition:
        return self.registry.get(name)

    def all(self) -> list[ConsentDefinition]:
        return sorted(list(self.registry.values()), key=lambda x: x.version)

    def validate_updates_or_raise(self, cdef: ConsentDefinition) -> None:
        if cdef.updates:
            if cdef.updates not in self.registry.values():
                raise ConsentDefinitionError(
                    f"Updates unregistered consent definition. See {cdef.name}. "
                    f"Got {cdef.updates.name}"
                )
            elif cdef.updates and cdef.updates.updated_by is None:
                raise ConsentDefinitionError(
                    f"Cdef mismatch with consent definition configured to update another. "
                    f"'{cdef.name}' is configured to update "
                    f"'{cdef.updates.name}' but '{cdef.updates.name}' "
                    f"updated_by is None. "
                )
            elif cdef.updates and cdef.updates.updated_by != cdef:
                raise ConsentDefinitionError(
                    f"Cdef mismatch with consent definition configured to update another. "
                    f"'{cdef.name}' is configured to update "
                    f"'{cdef.updates.name}' but '{cdef.updates.name}' "
                    f"updated_by='{cdef.updates.updated_by.name}' not '{cdef.name}'. "
                )

    def validate_period_overlap_or_raise(self, cdef: ConsentDefinition):
        for registered_cdef in self.registry.values():
            if (
                cdef
                and cdef.validate_duration_overlap_by_model
                and registered_cdef.proxy_model == cdef.proxy_model
            ):
                if (
                    registered_cdef.start <= cdef.start <= registered_cdef.end
                    or registered_cdef.start <= cdef.end <= registered_cdef.end
                ):
                    raise ConsentDefinitionError(
                        f"Consent period overlaps with an already registered consent "
                        f"definition. See already registered consent {registered_cdef.name}. "
                        f"Got {cdef.name}."
                    )

    def get_consents(self, subject_identifier: str, site_id: int | None) -> list:
        consents = []
        opts = {}
        if site_id:
            single_site = site_sites.get(site_id)
            opts.update(site=single_site)
        for cdef in self.get_consent_definitions(**opts):
            if consent_obj := cdef.get_consent_for(
                subject_identifier=subject_identifier,
                site_id=site_id,
                raise_if_not_consented=False,
            ):
                consents.append(consent_obj)
        return consents

    def get_consent_or_raise(
        self,
        subject_identifier: str,
        report_datetime: datetime,
        site_id: int | None = None,
        raise_if_not_consented: bool | None = None,
    ) -> ConsentLikeModel:
        """Returns a subject consent using this consent_definition's
        `model_cls` and `version`.

        If it does not exist and this consent_definition updates a
        previous (`update_cdef`), will try again with the `update_cdef's`
        model_cls and version.

        Finally, if the subject consent does not exist raises a
        `NotConsentedError`.
        """
        from edc_sites.site import sites as site_sites  # avoid circular import

        raise_if_not_consented = (
            True if raise_if_not_consented is None else raise_if_not_consented
        )

        single_site = site_sites.get(site_id) if site_id else None
        cdef = self.get_consent_definition(report_datetime=report_datetime, site=single_site)
        consent_obj = cdef.get_consent_for(
            subject_identifier=subject_identifier,
            raise_if_not_consented=raise_if_not_consented,
        )
        if consent_obj and to_utc(report_datetime) < consent_obj.consent_datetime:
            if not cdef.updates:
                dte = formatted_date(report_datetime)
                raise ConsentDefinitionNotConfiguredForUpdate(
                    f"Consent not configured to update any previous versions. "
                    f"Got '{cdef.version}'. "
                    f"Has subject '{subject_identifier}' completed version '{cdef.version}' "
                    f"of consent on or after report_datetime='{dte}'?"
                )
            else:
                return cdef.updates.get_consent_for(
                    subject_identifier, raise_if_not_consented=raise_if_not_consented
                )
        return consent_obj

    def get_consent_definition(
        self,
        model: str = None,
        report_datetime: datetime | None = None,
        version: str | None = None,
        site: SingleSite | None = None,
        screening_model: str | None = None,
        **kwargs,
    ) -> ConsentDefinition:
        """Returns a single consent definition valid for the given criteria.

        Filters the registry by each param given.
        """
        opts = dict(
            model=model,
            report_datetime=report_datetime,
            version=version,
            site=site,
            screening_model=screening_model,
        )
        cdefs = self.get_consent_definitions(**opts, **kwargs)
        if len(cdefs) > 1:
            as_string = ", ".join(list(set([cdef.name for cdef in cdefs])))
            raise SiteConsentError(
                f"Multiple consent definitions returned. Using {opts}. Got {as_string}. "
            )
        return cdefs[0]

    def get_consent_definitions(
        self,
        model: str = None,
        report_datetime: datetime | None = None,
        version: str | None = None,
        site: SingleSite | None = None,
        screening_model: str | None = None,
        **kwargs,
    ) -> list[ConsentDefinition]:
        """Return a list of consent definitions valid for the given
        criteria.

        Filters the registry by each param given.
        """
        error_messages: list[str] = []
        # confirm loaded
        if not self.registry.values() or not self.loaded:
            raise SiteConsentError(
                "No consent definitions have been registered with `site_consents`. "
            )
        # copy registry
        cdefs: list[ConsentDefinition] = [cdef for cdef in self.registry.values()]
        # filter cdefs to try to get just one.
        # by model, report_datetime, version, site
        cdefs, error_msg = self._filter_cdefs_by_model_or_raise(model, cdefs, error_messages)
        cdefs, error_msg = self._filter_cdefs_by_report_datetime_or_raise(
            report_datetime, cdefs, error_messages
        )
        cdefs, error_msg = self._filter_cdefs_by_version_or_raise(
            version, cdefs, error_messages
        )
        cdefs = self.filter_cdefs_by_site_or_raise(site, cdefs, error_messages)

        cdefs, error_msg = self._filter_cdefs_by_model_or_raise(
            screening_model, cdefs, error_messages, attrname="screening_model"
        )

        # apply additional criteria
        for k, v in kwargs.items():
            if v is not None:
                cdefs = [cdef for cdef in cdefs if getattr(cdef, k) == v]
        return sorted(cdefs, key=lambda x: x.version)

    @staticmethod
    def _filter_cdefs_by_model_or_raise(
        model: str | None,
        consent_definitions: list[ConsentDefinition],
        errror_messages: list[str] = None,
        attrname: str | None = None,
    ) -> tuple[list[ConsentDefinition], list[str]]:
        attrname = attrname or "model"
        cdefs = consent_definitions
        if model:
            cdefs = [cdef for cdef in cdefs if model == getattr(cdef, attrname)]
            if not cdefs:
                raise ConsentDefinitionDoesNotExist(
                    f"There are no consent definitions using this model. Got {model}."
                )
            else:
                errror_messages.append(f"model={model}")
        return cdefs, errror_messages

    def _filter_cdefs_by_report_datetime_or_raise(
        self,
        report_datetime: datetime | None,
        consent_definitions: list[ConsentDefinition],
        errror_messages: list[str] = None,
    ) -> tuple[list[ConsentDefinition], list[str]]:
        cdefs = consent_definitions
        if report_datetime:
            cdefs = [
                cdef
                for cdef in cdefs
                if floor_secs(cdef.start) <= to_utc(report_datetime) <= ceil_secs(cdef.end)
            ]
            if not cdefs:
                date_string = formatted_date(report_datetime)
                using_msg = "Using " + " and ".join(errror_messages)
                raise ConsentDefinitionDoesNotExist(
                    "Date does not fall within the validity period of any consent definition. "
                    f"Got {date_string}. {using_msg}. Consent definitions are: "
                    f"{self.get_registry_display()}."
                )
            else:
                date_string = formatted_date(report_datetime)
                errror_messages.append(f"report_datetime={date_string}")
        return cdefs, errror_messages

    def _filter_cdefs_by_version_or_raise(
        self,
        version: str | None,
        consent_definitions: list[ConsentDefinition],
        errror_messages: list[str] = None,
    ) -> tuple[list[ConsentDefinition], list[str]]:
        cdefs = consent_definitions
        if version:
            cdefs = [cdef for cdef in cdefs if cdef.version == version]
            if not cdefs:
                using_msg = "Using " + " and ".join(errror_messages)
                errror_messages.append(f"version={version}")
                raise ConsentDefinitionDoesNotExist(
                    f"There are no consent definitions for this version. "
                    f"Got {version}. {using_msg}. "
                    f"Consent definitions are: {self.get_registry_display()}."
                )
        return cdefs, errror_messages

    def filter_cdefs_by_site_or_raise(
        self,
        site: SingleSite | None,
        consent_definitions: list[ConsentDefinition],
        errror_messages: list[str] = None,
    ) -> list[ConsentDefinition]:
        cdefs = consent_definitions
        if site:
            cdefs_copy = [cdef for cdef in consent_definitions]
            cdefs = []
            for cdef in cdefs_copy:
                if site.site_id in [s.site_id for s in cdef.sites]:
                    cdefs.append(cdef)
            if not cdefs:
                using_msg = "Using " + " and ".join(errror_messages)
                raise ConsentDefinitionDoesNotExist(
                    f"There are no consent definitions for this site. "
                    f"Got {site}. {using_msg}."
                    f"Consent definitions are: {self.get_registry_display()}."
                )
        return cdefs

    def versions(self):
        return [cdef.version for cdef in self.registry.values()]

    def autodiscover(self, module_name=None, verbose=True):
        """Autodiscovers consent classes in the consents.py file of
        any INSTALLED_APP.
        """
        before_import_registry = None
        module_name = module_name or "consents"
        writer = sys.stdout.write if verbose else lambda x: x
        style = color_style()
        writer(f" * checking for site {module_name} ...\n")
        for app in django_apps.app_configs:
            writer(f" * searching {app}           \r")
            try:
                mod = import_module(app)
                try:
                    before_import_registry = deepcopy(site_consents.registry)
                    import_module(f"{app}.{module_name}")
                    writer(f" * registered consent definitions '{module_name}' from '{app}'\n")
                except SiteConsentError as e:
                    writer(f"   - loading {app}.consents ... ")
                    writer(style.ERROR(f"ERROR! {e}\n"))
                except ImportError as e:
                    site_consents.registry = before_import_registry
                    if module_has_submodule(mod, module_name):
                        raise SiteConsentError(str(e))
            except ImportError:
                pass
        for cdef in self.registry.values():
            start = cdef.start.strftime("%Y-%m-%d %Z")
            end = cdef.end.strftime("%Y-%m-%d %Z")
            sys.stdout.write(f"   - {cdef.name} valid {start} to {end}\n")


site_consents = SiteConsents()



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any

from django import forms
from django.apps import apps as django_apps
from django.conf import settings
from django.db import models
from edc_sites import site_sites

from .exceptions import ConsentDefinitionDoesNotExist
from .site_consents import site_consents

if TYPE_CHECKING:
    from edc_model.models import BaseUuidModel

    from edc_consent.consent_definition import ConsentDefinition

    from .model_mixins import ConsentModelMixin

    class ConsentModel(ConsentModelMixin, BaseUuidModel): ...


class InvalidInitials(Exception):
    pass


class MinimumConsentAgeError(Exception):
    pass


def get_consent_model_name() -> str:
    return settings.SUBJECT_CONSENT_MODEL


def get_consent_model_cls() -> Any:
    return django_apps.get_model(get_consent_model_name())


def get_consent_definition_or_raise(
    model: str | None = None,
    report_datetime: datetime | None = None,
    site_id: int | None = None,
    version: int | None = None,
) -> ConsentDefinition:
    opts = {}
    if model:
        opts.update(model=model)
    if report_datetime:
        opts.update(report_datetime=report_datetime)
    if version:
        opts.update(version=version)
    if site_id:
        opts.update(site=site_sites.get(site_id))
    try:
        consent_definition = site_consents.get_consent_definition(**opts)
    except ConsentDefinitionDoesNotExist as e:
        raise forms.ValidationError(e)
    return consent_definition


def get_reconsent_model_name() -> str:
    return getattr(
        settings,
        "SUBJECT_RECONSENT_MODEL",
        f"{get_consent_model_name().split('.')[0]}.subjectreconsent",
    )


def get_reconsent_model_cls() -> models.Model:
    return django_apps.get_model(get_reconsent_model_name())


def verify_initials_against_full_name(
    first_name: str | None = None,
    last_name: str | None = None,
    initials: str | None = None,
    **kwargs,  # noqa
) -> None:
    if first_name and initials and last_name:
        try:
            if initials[:1] != first_name[:1] or initials[-1:] != last_name[:1]:
                raise InvalidInitials("Initials do not match full name.")
        except (IndexError, TypeError):
            raise InvalidInitials("Initials do not match full name.")


def values_as_string(*values) -> str | None:
    if not any([True for v in values if v is None]):
        as_string = ""
        for value in values:
            try:
                value = value.isoformat()
            except AttributeError:
                pass
            as_string = f"{as_string}{value}"
        return as_string
    return None


def get_remove_patient_names_from_countries() -> list[str]:
    """Returns a list of country names."""
    return getattr(settings, "EDC_CONSENT_REMOVE_PATIENT_NAMES_FROM_COUNTRIES", [])



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from django.contrib import messages
from django.contrib.messages import ERROR
from edc_sites import site_sites

from .. import site_consents
from ..exceptions import ConsentDefinitionDoesNotExist, NotConsentedError

if TYPE_CHECKING:
    from django.db.models import QuerySet

    from ..consent_definition import ConsentDefinition
    from ..stubs import ConsentLikeModel


class ConsentViewMixin:
    consent_model: str | dict[str, str] = None

    """Declare with edc_appointment view mixin to get `appointment`."""

    def __init__(self, **kwargs):
        self._consent: ConsentLikeModel | None = None
        self._consents: QuerySet | None = None
        self._consent_definition: ConsentDefinition | None = None
        super().__init__(**kwargs)

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        """Add consent_definition and consents to the dashboard."""
        # TODO: What if active on more than one schedule??
        try:
            kwargs.update(consent_definition=self.consent_definition)
        except ConsentDefinitionDoesNotExist as e:
            messages.add_message(self.request, message=str(e), level=ERROR)
        else:
            kwargs.update(consent=self.consent, consents=self.consents)
        return super().get_context_data(**kwargs)

    @property
    def consents(self) -> QuerySet[ConsentLikeModel]:
        """Returns a Queryset of consents for this subject."""
        if not self._consents:
            self._consents = site_consents.get_consents(
                self.subject_identifier, site_id=self.request.site.id
            )
        return self._consents

    @property
    def consent(self) -> ConsentLikeModel | None:
        """Returns a consent model instance or None for the current
        period.
        """
        if not self._consent:
            try:
                self._consent = site_consents.get_consent_or_raise(
                    subject_identifier=self.subject_identifier,
                    report_datetime=self.report_datetime,
                    site_id=self.request.site.id,
                )
            except NotConsentedError as e:
                messages.add_message(self.request, message=str(e), level=ERROR)
        return self._consent

    @property
    def consent_definition(self) -> ConsentDefinition:
        """Returns a ConsentDefinition from the schedule for the
        current reporting period.
        """
        if not self._consent_definition:
            if self.current_schedule:
                self._consent_definition = self.current_schedule.get_consent_definition(
                    report_datetime=self.report_datetime,
                    site=site_sites.get(self.request.site.id),
                )
            elif self.appointment:
                self._consent_definition = self.appointment.schedule.get_consent_definition(
                    report_datetime=self.appointment.appt_datetime,
                    site=site_sites.get(self.appointment.site.id),
                )
        return self._consent_definition
